#include "testlib.h"
#include <vector>
#include <algorithm>
#include <numeric>
#include <queue>
#include <cassert>
#include <functional>

using namespace std;

namespace TreeGenerator {
  vector<int> generate_k_bamboo(int n, int k) {
    assert(n - 1 >= k);
    vector<int> parent(n, -1);
    vector<int> chain_sizes = rnd.partition(k, n - 1, 1);
    int cur = 1;
    for (int chain = 0; chain < k; chain++) {
      int size = chain_sizes[chain];
      parent[cur] = 0;
      for (int j = 1; j < size; j++) {
        parent[cur + j] = cur + j - 1;
      }
      cur += size;
    }
    return parent;
  }

  vector<int> generate_perfect_binary(int n) {
    vector<int> parent(n, -1);
    for (int i = 1; i < n; i++) {
      parent[i] = (i - 1) / 2;
    }
    return parent;
  }

  vector<int> generate_bamboo_brush_tree(int n, int max_h = -1) {
    if (max_h == -1) {
      max_h = n / 3; 
    }
    assert(max_h < n); 
    vector<int> parent(n, -1);
    for (int i = 1; i <= max_h; ++i) {
      parent[i] = i - 1;
    }
    for (int i = max_h + 1; i < n; ++i) {
      parent[i] = rnd.next(max_h);
    }
    return parent; 
  }

  vector<int> generate_random_tree_prufer(int n) {
    if (n == 1) {
      return vector<int>{-1};
    }
    vector<int> prufer(n - 2);
    for (int i = 0; i < n - 2; i++) {
      prufer[i] = rnd.next(n);
    }
    vector<int> degree(n, 1);
    for (int v : prufer) {
      degree[v]++;
    }
    
    set<int> leaves;
    for (int i = 0; i < n; i++) {
      if (degree[i] == 1)
        leaves.insert(i);
    }
    
    vector<vector<int>> adj(n);
    for (int v : prufer) {
      int leaf = *leaves.begin();
      leaves.erase(leaves.begin());
      
      adj[leaf].push_back(v);
      adj[v].push_back(leaf);
      
      degree[v]--;
      if (degree[v] == 1)
        leaves.insert(v);
    }
    
    int u = *leaves.begin();
    leaves.erase(leaves.begin());
    int v = *leaves.begin();
    adj[u].push_back(v);
    adj[v].push_back(u);
    
    vector<int> parent(n, -1);
    function<void(int, int)> dfs = [&](int node, int par) {
      parent[node] = par;
      for (int nxt : adj[node]) {
        if (nxt == par)
          continue;
        dfs(nxt, node);
      }
    };
    
    dfs(0, -1);
    return parent;
  }

  vector<int> generate_random_tree_last_k(int n, int k, int max_h = -1) {
    vector<int> parent(n, -1);
    vector<int> depth(n, 0);
    deque<int> pool = {0};
    for (int i = 1; i < n; i++) {
      if (!pool.empty()) {
        int idx = rnd.next(pool.size());
        int par = pool[idx];
        parent[i] = par;
        depth[i] = depth[par] + 1;
        if (pool.front() == i - k) {
          pool.pop_front();
        }
        if (max_h == -1 || depth[i] < max_h) {
          pool.push_back(i);
        }
      } else {
        assert(false);
      }
    }
    return parent;
  }

  vector<int> generate_preferential_attachment(int n, int max_h = -1) {
    vector<int> parent(n, -1);
    vector<int> pool = {0};
    vector<int> degree(n, 0);
    vector<int> depth(n, 0);
    for (int i = 1; i < n; i++) {
      int idx = rnd.next(pool.size());
      int par = pool[idx];
      parent[i] = par;
      depth[i] = depth[par] + 1;
      degree[par]++;
      pool.push_back(par);
      pool.push_back(par);
      if (max_h == -1 || depth[i] < max_h) {
        pool.push_back(i);
      }
    }
    return parent;
  }

  vector<int> shuffle_tree(const vector<int>& parent) {
    int n = parent.size();
    vector<vector<int>> children(n);
    for (int i = 1; i < n; i++) {
      children[parent[i]].push_back(i);
    }
    vector<int> order;
    order.reserve(n);
    function<void(int)> dfs = [&](int cur) {
      shuffle(children[cur].begin(), children[cur].end());
      order.push_back(cur);
      for (int child : children[cur]) {
        dfs(child);
      }
    };
    dfs(0);
    vector<int> new_label(n);
    for (int i = 0; i < n; i++) {
      new_label[order[i]] = i;
    }
    vector<int> new_parent(n, -1);
    new_parent[0] = -1;
    for (int i = 1; i < n; i++) {
      new_parent[i] = new_label[parent[order[i]]];
    }
    return new_parent;
  }
};

namespace SkillsGenerator {
  vector<int> generate_min_heap(const vector<int>& parents) {
    int n = parents.size();
    vector<int> skills(n, -1);
    vector<vector<int>> children(n);
    for (int i = 1; i < n; ++i)
      children[parents[i]].push_back(i);
    vector<int> heap_roots;
    heap_roots.reserve(n);
    heap_roots.push_back(0);
    for (int skill_ptr = 0; skill_ptr < n; ++skill_ptr) {
      assert(!heap_roots.empty());
      int heap_roots_idx = rnd.next((int)heap_roots.size());
      int v = heap_roots[heap_roots_idx];
      skills[v] = skill_ptr;
      swap(heap_roots[heap_roots_idx], heap_roots.back());
      heap_roots.pop_back();
      for (int c : children[v])
        heap_roots.push_back(c);
    }
    return skills;
  }

  vector<int> generate_random(const vector<int>& parents) {
    int n = parents.size();
    return rnd.perm(n);
  }

  vector<int> generate_dfs_order(const vector<int>& parents, bool perturb = false) {
    int n = parents.size();
    vector<vector<int>> children(n);
    for (int i = 1; i < n; ++i)
      children[parents[i]].push_back(i);
    vector<int> order;
    order.reserve(n);
    function<void(int)> dfs = [&](int cur) {
      order.push_back(cur);
      for (int child : children[cur])
        dfs(child);
    };
    dfs(0);
    if (perturb) {
      for (int i = 0; i + 1 < (int)order.size(); i++) {
        if (rnd.next(10) == 0) {
          swap(order[i], order[i + 1]);
        }
      }
    }
    vector<int> skills(n, -1);
    int value = n - 1;
    for (int v : order)
      skills[v] = value--;
    return skills;
  }

  vector<int> generate_bfs_order(const vector<int>& parents, bool perturb = false) {
    int n = parents.size();
    vector<vector<int>> children(n);
    for (int i = 1; i < n; ++i)
      children[parents[i]].push_back(i);
    vector<int> order;
    order.reserve(n);
    queue<int> q;
    q.push(0);
    while (!q.empty()) {
      int cur = q.front();
      q.pop();
      order.push_back(cur);
      for (int child : children[cur])
        q.push(child);
    }
    if (perturb) {
      for (int i = 0; i + 1 < (int)order.size(); i++) {
        if (rnd.next(10) == 0) {
          swap(order[i], order[i + 1]);
        }
      }
    }
    vector<int> skills(n, -1);
    int value = n - 1;
    for (int v : order)
      skills[v] = value--;
    return skills;
  }
}

namespace QueryGenerator {
  vector<int> generate_random(int q, const vector<int>& parents, const vector<int>& skills) {
    int n = parents.size();
    vector<int> queries = rnd.distinct(q, n);
    return queries;
  }

  vector<int> generate_smallest(int q, const vector<int>& parents, const vector<int>& skills) {
    int n = parents.size();
    vector<int> order(n);
    iota(order.begin(), order.end(), 0);
    sort(order.begin(), order.end(), [&](int i, int j) {
      return skills[i] < skills[j];
    });
    vector<int> queries;
    queries.reserve(q);
    for (int i = 0; i < q; i++) {
      queries.push_back(order[i]);
    }
    shuffle(queries.begin(), queries.end());
    return queries;
  }
};

int main(int argc, char** argv) {
  registerGen(argc, argv, 1);
  int n = opt<int>("n");
  int q = opt<int>("q");
  assert(q <= n); 

  string tree_type = opt<string>("tree_type");
  bool shuffle = opt<bool>("shuffle", false);

  vector<int> parents;
  if (tree_type == "k_bamboo") {
    int k = opt<int>("k");
    parents = TreeGenerator::generate_k_bamboo(n, k);
  } else if (tree_type == "bamboo_brush") {
    int max_h = opt<int>("max_h", -1);
    parents = TreeGenerator::generate_bamboo_brush_tree(n, max_h);
  } else if (tree_type == "perfect_binary") {
    parents = TreeGenerator::generate_perfect_binary(n);
  } else if (tree_type == "random_tree_prufer") {
    parents = TreeGenerator::generate_random_tree_prufer(n);
  } else if (tree_type == "random_tree_last_k") {
    int k = opt<int>("k");
    int max_h = opt<int>("max_h", -1);
    parents = TreeGenerator::generate_random_tree_last_k(n, k, max_h);
  } else if (tree_type == "preferential_attachment") {
    int max_h = opt<int>("max_h", -1);
    parents = TreeGenerator::generate_preferential_attachment(n, max_h);
  } else {
    assert(false);
  }

  if (shuffle) {
    parents = TreeGenerator::shuffle_tree(parents);
  }

  string skill_order = opt<string>("skill_order");
  vector<int> skills;
  if (skill_order == "min_heap") {
    skills = SkillsGenerator::generate_min_heap(parents);
  } else if (skill_order == "random") {
    skills = SkillsGenerator::generate_random(parents);
  } else if (skill_order == "dfs_order") {
    bool perturb = opt<bool>("perturb");
    skills = SkillsGenerator::generate_dfs_order(parents, perturb);
  } else if (skill_order == "bfs_order") {
    bool perturb = opt<bool>("perturb");
    skills = SkillsGenerator::generate_bfs_order(parents, perturb);
  } else {
    assert(false);
  }

  string query_order = opt<string>("queries_type");
  vector<int> queries;
  if (query_order == "random") {
    queries = QueryGenerator::generate_random(q, parents, skills);
  } else if (query_order == "smallest") {
    queries = QueryGenerator::generate_smallest(q, parents, skills);
  } else {
    assert(false);
  }

  cout << n << ' ' << q << '\n';
  for (int i = 1; i < n; ++i) {
    cout << parents[i] + 1; 
    if (i + 1 < n) {
      cout << ' ';
    }
  }
  cout << '\n';
  for (int i = 0; i < n; ++i) {
    cout << skills[i] + 1;
    if (i + 1 < n) {
      cout << ' ';
    }
  }
  cout << '\n';
  for (int i = 0; i < q; ++i) {
    cout << queries[i] + 1;
    if (i + 1 < q) {
      cout << ' ';
    }
  }
  cout << '\n';
}
